@@ -2,39 +2,43 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file
 
-/**
- * Foreign Function Interface for interoperability with the C programming language.
- *
- * For further details, please see: https://dart.dev/server/c-interop
- *
- * {@category VM}
- */
+/// Foreign Function Interface for interoperability with the C programming language.
+///
+/// For further details, please see: https://dart.dev/server/c-interop.
+///
+/// {@category VM}
+@Since('2.6')
 library dart.ffi;
 
+import 'dart:_internal' show Since;
 import 'dart:isolate';
 import 'dart:typed_data';
 
-part "native_type.dart";
-part "allocation.dart";
-part "annotations.dart";
-part "dynamic_library.dart";
-part "struct.dart";
+part 'abi.dart';
+part 'abi_specific.dart';
+part 'native_type.dart';
+part 'native_finalizer.dart';
+part 'allocation.dart';
+part 'annotations.dart';
+part 'c_type.dart';
+part 'dynamic_library.dart';
+part 'struct.dart';
+part 'union.dart';
 
 /// Number of bytes used by native type T.
 ///
 /// Includes padding and alignment of structs.
 ///
-/// Support for invoking this function with non-constant [T] will be removed in
-/// the next stable version of Dart and it will become mandatory to invoke it
-/// with a compile-time constant [T].
+/// This function must be invoked with a compile-time constant [T].
 external int sizeOf<T extends NativeType>();
 
-/// Represents a pointer into the native C memory corresponding to "NULL", e.g.
+/// Represents a pointer into the native C memory corresponding to 'NULL', e.g.
 /// a pointer with address 0.
 final Pointer<Never> nullptr = Pointer.fromAddress(0);
 
 /// Represents a pointer into the native C memory. Cannot be extended.
-@pragma("vm:entry-point")
+@pragma('vm:entry-point')
+@pragma("wasm:entry-point")
 class Pointer<T extends NativeType> extends NativeType {
   /// Construction from raw integer.
   external factory Pointer.fromAddress(int ptr);
@@ -56,7 +60,7 @@ class Pointer<T extends NativeType> extends NativeType {
   /// Does not accept dynamic invocations -- where the type of the receiver is
   /// [dynamic].
   external static Pointer<NativeFunction<T>> fromFunction<T extends Function>(
-      @DartRepresentationOf("T") Function f,
+      @DartRepresentationOf('T') Function f,
       [Object? exceptionalReturn]);
 
   /// Access to the raw pointer value.
@@ -65,9 +69,10 @@ class Pointer<T extends NativeType> extends NativeType {
 
   /// Pointer arithmetic (takes element size into account).
   ///
-  /// Support for invoking this method with non-constant [T] will be removed in
-  /// the next stable version of Dart and it will become mandatory to invoke it
-  /// with a compile-time constant [T].
+  /// This method must be invoked with a compile-time constant [T].
+  ///
+  /// Does not accept dynamic invocations -- where the type of the receiver is
+  /// [dynamic].
   external Pointer<T> elementAt(int index);
 
   /// Cast Pointer<T> to a Pointer<V>.
@@ -86,12 +91,76 @@ class Pointer<T extends NativeType> extends NativeType {
   }
 }
 
+/// A fixed-sized array of [T]s.
+@Since('2.13')
+class Array<T extends NativeType> extends NativeType {
+  /// Const constructor to specify [Array] dimensions in [Struct]s.
+  ///
+  /// ```dart
+  /// class MyStruct extends Struct {
+  ///   @Array(8)
+  ///   external Array<Uint8> inlineArray;
+  ///
+  ///   @Array(2, 2, 2)
+  ///   external Array<Array<Array<Uint8>>> threeDimensionalInlineArray;
+  /// }
+  /// ```
+  ///
+  /// Do not invoke in normal code.
+  const factory Array(int dimension1,
+      [int dimension2,
+      int dimension3,
+      int dimension4,
+      int dimension5]) = _ArraySize<T>;
+
+  /// Const constructor to specify [Array] dimensions in [Struct]s.
+  ///
+  /// ```dart
+  /// class MyStruct extends Struct {
+  ///   @Array.multi([2, 2, 2])
+  ///   external Array<Array<Array<Uint8>>> threeDimensionalInlineArray;
+  ///
+  ///   @Array.multi([2, 2, 2, 2, 2, 2, 2, 2])
+  ///   external Array<Array<Array<Array<Array<Array<Array<Array<Uint8>>>>>>>> eightDimensionalInlineArray;
+  /// }
+  /// ```
+  ///
+  /// Do not invoke in normal code.
+  const factory Array.multi(List<int> dimensions) = _ArraySize<T>.multi;
+}
+
+class _ArraySize<T extends NativeType> implements Array<T> {
+  final int? dimension1;
+  final int? dimension2;
+  final int? dimension3;
+  final int? dimension4;
+  final int? dimension5;
+
+  final List<int>? dimensions;
+
+  const _ArraySize(this.dimension1,
+      [this.dimension2, this.dimension3, this.dimension4, this.dimension5])
+      : dimensions = null;
+
+  const _ArraySize.multi(this.dimensions)
+      : dimension1 = null,
+        dimension2 = null,
+        dimension3 = null,
+        dimension4 = null,
+        dimension5 = null;
+}
+
 /// Extension on [Pointer] specialized for the type argument [NativeFunction].
 extension NativeFunctionPointer<NF extends Function>
     on Pointer<NativeFunction<NF>> {
   /// Convert to Dart function, automatically marshalling the arguments
   /// and return value.
-  external DF asFunction<@DartRepresentationOf("NF") DF extends Function>();
+  ///
+  /// [isLeaf] specifies whether the function is a leaf function.
+  /// A leaf function must not run Dart code or call back into the Dart VM.
+  /// Leaf calls are faster than non-leaf calls.
+  external DF asFunction<@DartRepresentationOf('NF') DF extends Function>(
+      {bool isLeaf = false});
 }
 
 //
@@ -386,50 +455,6 @@ extension Uint64Pointer on Pointer<Uint64> {
   external Uint64List asTypedList(int length);
 }
 
-/// Extension on [Pointer] specialized for the type argument [IntPtr].
-extension IntPtrPointer on Pointer<IntPtr> {
-  /// The 32 or 64-bit two's complement integer at [address].
-  ///
-  /// On 32-bit platforms this is a 32-bit integer, and on 64-bit platforms
-  /// this is a 64-bit integer.
-  ///
-  /// On 32-bit platforms a Dart integer is truncated to 32 bits (as if by
-  /// `.toSigned(32)`) before being stored, and the 32-bit value is
-  /// sign-extended when it is loaded.
-  ///
-  /// On 32-bit platforms the [address] must be 4-byte aligned, and on 64-bit
-  /// platforms the [address] must be 8-byte aligned.
-  external int get value;
-
-  external void set value(int value);
-
-  /// The 32 or 64-bit two's complement integer at `address + (4 or 8) * index`.
-  ///
-  /// On 32-bit platforms this is a 32-bit integer, and on 64-bit platforms
-  /// this is a 64-bit integer.
-  ///
-  /// On 32-bit platforms a Dart integer is truncated to 32 bits (as if by
-  /// `.toSigned(32)`) before being stored, and the 32-bit value is
-  /// sign-extended when it is loaded.
-  ///
-  /// On 32-bit platforms the [address] must be 4-byte aligned, and on 64-bit
-  /// platforms the [address] must be 8-byte aligned.
-  external int operator [](int index);
-
-  /// The 32 or 64-bit two's complement integer at `address + (4 or 8) * index`.
-  ///
-  /// On 32-bit platforms this is a 32-bit integer, and on 64-bit platforms
-  /// this is a 64-bit integer.
-  ///
-  /// On 32-bit platforms a Dart integer is truncated to 32 bits (as if by
-  /// `.toSigned(32)`) before being stored, and the 32-bit value is
-  /// sign-extended when it is loaded.
-  ///
-  /// On 32-bit platforms the [address] must be 4-byte aligned, and on 64-bit
-  /// platforms the [address] must be 8-byte aligned.
-  external void operator []=(int index, int value);
-}
-
 /// Extension on [Pointer] specialized for the type argument [Float].
 extension FloatPointer on Pointer<Float> {
   /// The float at [address].
@@ -501,6 +526,109 @@ extension DoublePointer on Pointer<Double> {
   external Float64List asTypedList(int length);
 }
 
+/// Extension on [Pointer] specialized for the type argument [Bool].
+@Since('2.15')
+extension BoolPointer on Pointer<Bool> {
+  /// The bool at [address].
+  external bool get value;
+
+  external void set value(bool value);
+
+  /// The bool at `address + index`.
+  external bool operator [](int index);
+
+  /// The bool at `address + index`.
+  external void operator []=(int index, bool value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Int8].
+@Since('2.13')
+extension Int8Array on Array<Int8> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Int16].
+@Since('2.13')
+extension Int16Array on Array<Int16> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Int32].
+@Since('2.13')
+extension Int32Array on Array<Int32> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Int64].
+@Since('2.13')
+extension Int64Array on Array<Int64> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Uint8].
+@Since('2.13')
+extension Uint8Array on Array<Uint8> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Uint16].
+@Since('2.13')
+extension Uint16Array on Array<Uint16> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Uint32].
+@Since('2.13')
+extension Uint32Array on Array<Uint32> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Uint64].
+@Since('2.13')
+extension Uint64Array on Array<Uint64> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Float].
+@Since('2.13')
+extension FloatArray on Array<Float> {
+  external double operator [](int index);
+
+  external void operator []=(int index, double value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Double].
+@Since('2.13')
+extension DoubleArray on Array<Double> {
+  external double operator [](int index);
+
+  external void operator []=(int index, double value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Bool].
+@Since('2.15')
+extension BoolArray on Array<Bool> {
+  external bool operator [](int index);
+
+  external void operator []=(int index, bool value);
+}
+
 //
 // End of generated code.
 //
@@ -530,7 +658,7 @@ extension PointerPointer<T extends NativeType> on Pointer<Pointer<T>> {
   /// Store a Dart value into this location offset by [index].
   ///
   /// A [Pointer] is unboxed before being stored (as if by `.address`), and the
-  /// pointer is boxed (as if by `Pointer.fromAddress`) when loaded.
+  /// pointer is boxed (as if by [Pointer.fromAddress]) when loaded.
   ///
   /// On 32-bit platforms the [address] must be 4-byte aligned, and on 64-bit
   /// platforms the [address] must be 8-byte aligned.
@@ -538,17 +666,22 @@ extension PointerPointer<T extends NativeType> on Pointer<Pointer<T>> {
 }
 
 /// Extension on [Pointer] specialized for the type argument [Struct].
+@Since('2.12')
 extension StructPointer<T extends Struct> on Pointer<T> {
-  /// Creates a reference to access the fields of this struct backed by native
-  /// memory at [address].
+  /// A Dart view of the struct referenced by this pointer.
   ///
+  /// Reading [ref] creates a reference accessing the fields of this struct
+  /// backed by native memory at [address].
   /// The [address] must be aligned according to the struct alignment rules of
   /// the platform.
   ///
-  /// Support for invoking this extension method with non-constant [T] will be
-  /// removed in the next stable version of Dart and it will become mandatory
-  /// to invoke it with a compile-time constant [T].
+  /// Assigning to [ref] copies contents of the struct into the native memory
+  /// starting at [address].
+  ///
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
   external T get ref;
+  external set ref(T value);
 
   /// Creates a reference to access the fields of this struct backed by native
   /// memory at `address + sizeOf<T>() * index`.
@@ -556,38 +689,146 @@ extension StructPointer<T extends Struct> on Pointer<T> {
   /// The [address] must be aligned according to the struct alignment rules of
   /// the platform.
   ///
-  /// Support for invoking this extension method with non-constant [T] will be
-  /// removed in the next stable version of Dart and it will become mandatory
-  /// to invoke it with a compile-time constant [T].
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external T operator [](int index);
+
+  /// Copies the [value] struct into native memory, starting at
+  /// `address * sizeOf<T>() * index`.
+  ///
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external void operator []=(int index, T value);
+}
+
+/// Extension on [Pointer] specialized for the type argument [Union].
+@Since('2.14')
+extension UnionPointer<T extends Union> on Pointer<T> {
+  /// A Dart view of the union referenced by this pointer.
+  ///
+  /// Reading [ref] creates a reference accessing the fields of this union
+  /// backed by native memory at [address].
+  /// The [address] must be aligned according to the union alignment rules of
+  /// the platform.
+  ///
+  /// Assigning to [ref] copies contents of the union into the native memory
+  /// starting at [address].
+  ///
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external T get ref;
+  external set ref(T value);
+
+  /// Creates a reference to access the fields of this union backed by native
+  /// memory at `address + sizeOf<T>() * index`.
+  ///
+  /// The [address] must be aligned according to the union alignment rules of
+  /// the platform.
+  ///
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external T operator [](int index);
+
+  /// Copies the [value] union into native memory, starting at
+  /// `address * sizeOf<T>() * index`.
+  ///
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external void operator []=(int index, T value);
+}
+
+/// Extension on [Pointer] specialized for the type argument
+/// [AbiSpecificInteger].
+@Since('2.16')
+extension AbiSpecificIntegerPointer<T extends AbiSpecificInteger>
+    on Pointer<T> {
+  /// The integer at [address].
+  external int get value;
+
+  external void set value(int value);
+
+  /// The integer at `address + sizeOf<T>() * index`.
+  external int operator [](int index);
+
+  /// The integer at `address + sizeOf<T>() * index`.
+  external void operator []=(int index, int value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Pointer].
+@Since('2.13')
+extension PointerArray<T extends NativeType> on Array<Pointer<T>> {
+  external Pointer<T> operator [](int index);
+
+  external void operator []=(int index, Pointer<T> value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Struct].
+@Since('2.13')
+extension StructArray<T extends Struct> on Array<T> {
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
   external T operator [](int index);
 }
 
+/// Bounds checking indexing methods on [Array]s of [Union].
+@Since('2.14')
+extension UnionArray<T extends Union> on Array<T> {
+  /// This extension method must be invoked on a receiver of type `Pointer<T>`
+  /// where `T` is a compile-time constant type.
+  external T operator [](int index);
+}
+
+/// Bounds checking indexing methods on [Array]s of [Array].
+@Since('2.13')
+extension ArrayArray<T extends NativeType> on Array<Array<T>> {
+  external Array<T> operator [](int index);
+
+  external void operator []=(int index, Array<T> value);
+}
+
+/// Bounds checking indexing methods on [Array]s of [AbiSpecificInteger].
+@Since('2.16')
+extension AbiSpecificIntegerArray on Array<AbiSpecificInteger> {
+  external int operator [](int index);
+
+  external void operator []=(int index, int value);
+}
+
 /// Extension to retrieve the native `Dart_Port` from a [SendPort].
+@Since('2.7')
 extension NativePort on SendPort {
   /// The native port of this [SendPort].
   ///
   /// The returned native port can for example be used by C code to post
   /// messages to the connected [ReceivePort] via `Dart_PostCObject()` - see
   /// `dart_native_api.h`.
+  ///
+  /// Only the send ports from the platform classes [ReceivePort] and
+  /// [RawReceivePort] are supported. User-defined implementations of
+  /// [SendPort] are not supported.
   external int get nativePort;
 }
 
 /// Opaque, not exposing it's members.
+@Since('2.8')
 class Dart_CObject extends Opaque {}
 
 typedef Dart_NativeMessageHandler = Void Function(Int64, Pointer<Dart_CObject>);
 
 /// Utilities for accessing the Dart VM API from Dart code or
 /// from C code via `dart_api_dl.h`.
+@Since('2.8')
 abstract class NativeApi {
   /// On breaking changes the major version is increased.
   ///
   /// The versioning covers the API surface in `dart_api_dl.h`.
+  @Since('2.9')
   external static int get majorVersion;
 
   /// On backwards compatible changes the minor version is increased.
   ///
   /// The versioning covers the API surface in `dart_api_dl.h`.
+  @Since('2.9')
   external static int get minorVersion;
 
   /// A function pointer to
@@ -598,7 +839,7 @@ abstract class NativeApi {
       get postCObject;
 
   /// A function pointer to
-  /// ```
+  /// ```c
   /// Dart_Port Dart_NewNativePort(const char* name,
   ///                              Dart_NativeMessageHandler handler,
   ///                              bool handle_concurrently)
@@ -619,5 +860,44 @@ abstract class NativeApi {
 
   /// Pass this to `Dart_InitializeApiDL` in your native code to enable using the
   /// symbols in `dart_api_dl.h`.
+  @Since('2.9')
   external static Pointer<Void> get initializeApiDLData;
 }
+
+/// Annotation to be used for marking an external function as FFI native.
+///
+/// Example:
+///```dart template:top
+/// @FfiNative<Int64 Function(Int64, Int64)>('FfiNative_Sum', isLeaf:true)
+/// external int sum(int a, int b);
+///```
+/// Calling such functions will throw an exception if no resolver
+/// was set on the library or the resolver failed to resolve the name.
+///
+/// See `Dart_SetFfiNativeResolver` in `dart_api.h`
+///
+/// NOTE: This is an experimental feature and may change in the future.
+@Since('2.14')
+class FfiNative<T> {
+  final String nativeName;
+
+  /// Specifies whether the function is a leaf function.
+  ///
+  /// A leaf function must not run Dart code or call back into the Dart VM.
+  /// Leaf calls are faster than non-leaf calls.
+  final bool isLeaf;
+
+  const FfiNative(this.nativeName, {this.isLeaf = false});
+}
+
+// Bootstrapping native for getting the FFI native C function pointer to look
+// up the FFI resolver.
+@pragma('vm:external-name', 'Ffi_GetFfiNativeResolver')
+external Pointer<NativeFunction<IntPtr Function(Handle, Handle, IntPtr)>>
+    _get_ffi_native_resolver<T extends NativeFunction>();
+
+// Resolver for FFI Native C function pointers.
+@pragma('vm:entry-point')
+final _ffi_resolver = _get_ffi_native_resolver<
+        NativeFunction<IntPtr Function(Handle, Handle, IntPtr)>>()
+    .asFunction<int Function(Object, Object, int)>();
